#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/io.h>

#include "jamexprt.h"

static int verbose;

static int port_list[] = { 0x378, 0x278, 0x3bc, 0x290, 0 };
static int port;

#undef inb
#undef outb
int inb (int port);
void outb (int port, int val);

asm ("
inb:
	movl 4(%esp),%edx
	xorl %eax,%eax
	inb %dx, %al
	ret

outb:
	movl 8(%esp),%eax
	movl 4(%esp),%edx
	outb %al, %dx
	ret

");

static int
par_init ()
{
	return (ioperm (port, 2, 1));
}

static int
par_input ()
{
	return (inb (port + 1));
}

static void
par_output (int val)
{
	outb (port, val);
}


static FILE *jamf;
static char *init_list[100];

static void
usage ()
{
	fprintf (stderr, "usage: jamp [opts] file.jam\n");
	fprintf (stderr, "  -p 0xNNN      io port (default 0x%x)\n",
		 port_list[0]);
	fprintf (stderr, "  -v            verbose\n");
	exit (1);
}

static char *
jamrc_str (int rc)
{
	static char buf[1000];

	switch (rc) {
	case JAMC_SUCCESS: return ("JAMC_SUCCESS");
	case JAMC_OUT_OF_MEMORY: return ("JAMC_OUT_OF_MEMORY");
	case JAMC_IO_ERROR: return ("JAMC_IO_ERROR");
	case JAMC_SYNTAX_ERROR: return ("JAMC_SYNTAX_ERROR");
	case JAMC_UNEXPECTED_END: return ("JAMC_UNEXPECTED_END");
	case JAMC_UNDEFINED_SYMBOL: return ("JAMC_UNDEFINED_SYMBOL");
	case JAMC_REDEFINED_SYMBOL: return ("JAMC_REDEFINED_SYMBOL");
	case JAMC_INTEGER_OVERFLOW: return ("JAMC_INTEGER_OVERFLOW");
	case JAMC_DIVIDE_BY_ZERO: return ("JAMC_DIVIDE_BY_ZERO");
	case JAMC_CRC_ERROR: return ("JAMC_CRC_ERROR");
	case JAMC_INTERNAL_ERROR: return ("JAMC_INTERNAL_ERROR");
	case JAMC_BOUNDS_ERROR: return ("JAMC_BOUNDS_ERROR");
	case JAMC_TYPE_MISMATCH: return ("JAMC_TYPE_MISMATCH");
	case JAMC_ASSIGN_TO_CONST: return ("JAMC_ASSIGN_TO_CONST");
	case JAMC_NEXT_UNEXPECTED: return ("JAMC_NEXT_UNEXPECTED");
	case JAMC_POP_UNEXPECTED: return ("JAMC_POP_UNEXPECTED");
	case JAMC_RETURN_UNEXPECTED: return ("JAMC_RETURN_UNEXPECTED");
	case JAMC_ILLEGAL_SYMBOL: return ("JAMC_ILLEGAL_SYMBOL");
	case JAMC_VECTOR_MAP_FAILED: return ("JAMC_VECTOR_MAP_FAILED");
	default:
		sprintf (buf, "JAM_RC_%d", rc);
		return (buf);
	}
}

int
main (int argc, char **argv)
{
	int c;
	char *filename;
	long error_line;
	int exit_code;
	int rc;
	long offset;
	char key[1000];
	char value[10000];
	int i;

	setbuf (stdout, NULL);

	port = 0;

	while ((c = getopt (argc, argv, "vp:")) != EOF) {
		switch (c) {
		case 'p':
			port = strtol (optarg, NULL, 0);
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			usage ();
		}
	}

	if (optind >= argc)
		usage ();

	filename = argv[optind++];

	if (optind != argc)
		usage ();

	if (port == 0) {
		port = port_list[0];
	} else if (getuid () != 0) {
		for (i = 0; port_list[i]; i++)
			if (port == port_list[i])
				break;
		if (port_list[i] == 0) {
			fprintf (stderr,
				 "illegal port number 0x%x\n",
				 port);
			fprintf (stderr,
				 "to use this port, you must run"
				 " as root\n");
			exit (1);
		}
	}
	
	if (par_init () < 0) {
		perror ("ioperm");
		if (geteuid () != 0)
			fprintf (stderr,
				 "jamp should be installed setuid root\n");
		exit (1);
	}

	setuid (getuid ());

	if ((jamf = fopen (filename, "r")) == NULL) {
		fprintf (stderr, "can't open %s\n", filename);
		exit (1);
	}

	offset = 0;
	while (1) {
		rc = jam_get_note (NULL, 0, &offset, key, value, sizeof value);
		if (rc != JAMC_SUCCESS)
			break;
		value[sizeof value - 1] = 0;
		printf ("NOTE \"%s\" = \"%s\"\n", key, value);
	}

	init_list[0] = NULL;
	exit_code = 0;
	rc = jam_execute (NULL, 0, NULL, 0, init_list, 
			  &error_line, &exit_code);

	if (rc != JAMC_SUCCESS) {
		printf ("%s:%ld: jam_execute error %s %d\n",
			filename, error_line, jamrc_str (rc), exit_code);
		exit (1);
	}

	return (0);
}

int
jam_getc (void)
{
	return (getc (jamf));
}

int
jam_seek (long offset)
{
	fseek (jamf, offset, SEEK_SET);
	return (0);
}

void
jam_export (char *key, long value)
{
	printf ("jam_export %s %ld\n", key, value);
}

void
jam_message (char *message_text)
{
	printf ("jam_message: %s\n", message_text);
}

static char **jam_vectors;
static int jam_nvectors;

int
jam_vector_map (int signal_count, char **signals)
{
	int i;

	if (jam_nvectors) {
		fprintf (stderr, "duplicate vector_map\n");
		exit (1);
	}
	jam_nvectors = signal_count;
	jam_vectors = jam_malloc (signal_count * sizeof *jam_vectors);
	for (i = 0; i < signal_count; i++) {
		if ((jam_vectors[i] = strdup (signals[i])) == NULL) {
			fprintf (stderr, "out of memory\n");
			exit (1);
		}
	}

	printf ("jam_vector_map:");
	for (i = 0; i < jam_nvectors; i++)
		printf ("  %d: %s\n", i, jam_vectors[i]);
	printf ("END\n");

	return (jam_nvectors);
}

int
jam_vector_io (int signal_count, long *dir_vect, long *data_vect, long *capture_vect)
{
	int i;

	if (signal_count > jam_nvectors) {
		fprintf (stderr, "out of range vector_io\n");
		exit (1);
	}

	printf ("VECTOR_IO\n");
	for (i = 0; i < signal_count; i++) {
		printf ("  %d: ", i);
		if (dir_vect)
			printf ("0x%lx ", dir_vect[i]);
		else
			printf ("(null) ");
		if (data_vect)
			printf ("0x%lx ", data_vect[i]);
		else
			printf ("(null) ");
		if (capture_vect)
			printf ("0x%lx ", capture_vect[i]);
		else
			printf ("(null) ");
		printf ("\n");
	}
	printf ("END\n");
	return (signal_count);
}

void
jam_free (void *ptr)
{
	free (ptr);
}

void *
jam_malloc (unsigned int size)
{
	void *p;

	if ((p = malloc (size)) == NULL) {
		fprintf (stderr, "out of memory\n");
		exit (1);
	}

	return (p);
}

/* this state stuff is just to help make the debugging output more readable */

enum jtag_state {
	js_reset,
	js_idle,
	js_drselect,
	js_drcapture,
	js_drshift,
	js_drexit1,
	js_drpause,
	js_drexit2,
	js_drupdate,
	js_irselect,
	js_ircapture,
	js_irshift,
	js_irexit1,
	js_irpause,
	js_irexit2,
	js_irupdate,
};

static enum jtag_state jtag_state;

struct jtag_next_state {
	enum jtag_state tms_lo;
	enum jtag_state tms_hi;
};

static struct jtag_next_state jtag_next_state_table[] = {
	/* 			tms=0		tms=1 */
	{ /* js_reset */	js_idle,	js_reset }, 
	{ /* js_idle */		js_idle,	js_drselect },
	{ /* js_drselect */	js_drcapture,	js_irselect },
	{ /* js_drcapture */	js_drshift,	js_drexit1 },
	{ /* js_drshift */	js_drshift,	js_drexit1 },
	{ /* js_drexit1 */	js_drpause,	js_drupdate },
	{ /* js_drpause */	js_drpause,	js_drexit2 },
	{ /* js_drexit2 */	js_drshift,	js_drupdate },
	{ /* js_drupdate */	js_idle,	js_drselect },
	{ /* js_irselect */	js_ircapture,	js_reset },
	{ /* js_ircapture */	js_irshift,	js_irexit1 },
	{ /* js_irshift */	js_irshift,	js_irexit1 },
	{ /* js_irexit1 */	js_irpause,	js_irupdate },
	{ /* js_irpause */	js_irpause,	js_irexit2 },
	{ /* js_irexit2 */	js_irshift,	js_irupdate },
	{ /* js_irupdate */	js_idle,	js_drselect },
};

static char *jtag_state_name[] = {
	"reset",
	"idle",
	"drselect",
	"drcapture",
	"drshift",
	"drexit1",
	"drpause",
	"drexit2",
	"drupdate",
	"irselect",
	"ircapture",
	"irshift",
	"irexit1",
	"irpause",
	"irexit2",
	"irupdate",
};
	
static enum jtag_state jtag_state;

int
jam_jtag_io (int tms, int tdi, int read_tdo)
{
	int retval;
	int val;

	val = 0x8; /* ISR */
	if (tms)
		val |= 4;
	if (tdi)
		val |= 2;
	par_output (val);
	/* XXX delay needed? */
	par_output (val | 1); /* TCK = 1 */
	/* XXX delay needed? */

	/* perhaps this should be at the very start of this function? */
	retval = 0;
	if (read_tdo)
		if (par_input () & 0x40)
			retval = 1;

	par_output (val | 0); /* TCK = 0 */

	if (tms)
		jtag_state = jtag_next_state_table[jtag_state].tms_hi;
	else
		jtag_state = jtag_next_state_table[jtag_state].tms_lo;

	if (verbose)
		printf ("jtag_io: %d %d => %d %s\n",
			tms, tdi, retval, jtag_state_name[jtag_state]);

	return (retval);
}

void
jam_delay (long microseconds)
{
	usleep (microseconds);
}

